;;; phps-mode-syntax-color.el --- Syntax coloring for major mode -*- lexical-binding: t -*-

;; Copyright (C) 2018-2023  Free Software Foundation, Inc.


;;; Commentary:


;;; Code:


(defvar
  phps-mode-syntax-color--token-font-face
  #s(hash-table size 151 test equal rehash-size 1.5 rehash-threshold 0.8125 data (T_ERROR font-lock-warning-face T_OPEN_TAG font-lock-constant-face T_OPEN_TAG_WITH_ECHO font-lock-constant-face T_CLOSE_TAG font-lock-constant-face T_START_HEREDOC font-lock-constant-face T_END_HEREDOC font-lock-constant-face T_ELLIPSIS font-lock-constant-face T_COALESCE font-lock-constant-face T_DOUBLE_ARROW font-lock-constant-face T_INC font-lock-constant-face T_DEC font-lock-constant-face T_IS_IDENTICAL font-lock-constant-face T_IS_NOT_IDENTICAL font-lock-constant-face T_IS_EQUAL font-lock-constant-face T_IS_NOT_EQUAL font-lock-constant-face T_SPACESHIP font-lock-constant-face T_IS_SMALLER_OR_EQUAL font-lock-constant-face T_IS_GREATER_OR_EQUAL font-lock-constant-face T_PLUS_EQUAL font-lock-constant-face T_MINUS_EQUAL font-lock-constant-face T_MUL_EQUAL font-lock-constant-face T_POW_EQUAL font-lock-constant-face T_POW font-lock-constant-face T_DIV_EQUAL font-lock-constant-face T_CONCAT_EQUAL font-lock-constant-face T_MOD_EQUAL font-lock-constant-face T_SL_EQUAL font-lock-constant-face T_SR_EQUAL font-lock-constant-face T_AND_EQUAL font-lock-constant-face T_OR_EQUAL font-lock-constant-face T_XOR_EQUAL font-lock-constant-face T_COALESCE_EQUAL font-lock-constant-face T_BOOLEAN_OR font-lock-constant-face T_BOOLEAN_AND font-lock-constant-face T_BOOLEAN_XOR font-lock-constant-face T_LOGICAL_XOR font-lock-constant-face T_LOGICAL_OR font-lock-constant-face T_LOGICAL_AND font-lock-constant-face T_SL font-lock-constant-face T_SR font-lock-constant-face T_CLASS_C font-lock-constant-face T_TRAIT_C font-lock-constant-face T_FUNC_C font-lock-constant-face T_METHOD_C font-lock-constant-face T_LINE font-lock-constant-face T_FILE font-lock-constant-face T_DIR font-lock-constant-face T_NS_C font-lock-constant-face T_INT_CAST font-lock-constant-face T_DOUBLE_CAST font-lock-constant-face T_STRING_CAST font-lock-constant-face T_ARRAY_CAST font-lock-constant-face T_OBJECT_CAST font-lock-constant-face T_BOOL_CAST font-lock-constant-face T_UNSET_CAST font-lock-constant-face T_DOLLAR_OPEN_CURLY_BRACES font-lock-keyword-face T_CURLY_OPEN font-lock-keyword-face T_OBJECT_OPERATOR font-lock-keyword-face T_NULLSAFE_OBJECT_OPERATOR font-lock-keyword-face T_PAAMAYIM_NEKUDOTAYIM font-lock-keyword-face T_NS_SEPARATOR font-lock-keyword-face T_EXIT font-lock-keyword-face T_RETURN font-lock-keyword-face T_YIELD_FROM font-lock-keyword-face T_YIELD font-lock-keyword-face T_TRY font-lock-keyword-face T_CATCH font-lock-keyword-face T_FINALLY font-lock-keyword-face T_THROW font-lock-keyword-face T_IF font-lock-keyword-face T_ELSEIF font-lock-keyword-face T_ENDIF font-lock-keyword-face T_ELSE font-lock-keyword-face T_WHILE font-lock-keyword-face T_ENDWHILE font-lock-keyword-face T_DO font-lock-keyword-face T_FUNCTION font-lock-keyword-face T_FN font-lock-keyword-face T_CONST font-lock-keyword-face T_FOREACH font-lock-keyword-face T_ENDFOREACH font-lock-keyword-face T_FOR font-lock-keyword-face T_ENDFOR font-lock-keyword-face T_DECLARE font-lock-keyword-face T_ENDDECLARE font-lock-keyword-face T_INSTANCEOF font-lock-keyword-face T_AS font-lock-keyword-face T_SWITCH font-lock-keyword-face T_MATCH font-lock-keyword-face T_ENDSWITCH font-lock-keyword-face T_CASE font-lock-keyword-face T_DEFAULT font-lock-keyword-face T_BREAK font-lock-keyword-face T_CONTINUE font-lock-keyword-face T_GOTO font-lock-keyword-face T_ECHO font-lock-keyword-face T_PRINT font-lock-keyword-face T_CLASS font-lock-keyword-face T_INTERFACE font-lock-keyword-face T_TRAIT font-lock-keyword-face T_EXTENDS font-lock-keyword-face T_IMPLEMENTS font-lock-keyword-face T_NEW font-lock-keyword-face T_CLONE font-lock-keyword-face T_VAR font-lock-keyword-face T_EVAL font-lock-keyword-face T_INCLUDE_ONCE font-lock-keyword-face T_INCLUDE font-lock-keyword-face T_REQUIRE_ONCE font-lock-keyword-face T_REQUIRE font-lock-keyword-face T_NAMESPACE font-lock-keyword-face T_USE font-lock-keyword-face T_INSTEADOF font-lock-keyword-face T_GLOBAL font-lock-keyword-face T_ISSET font-lock-keyword-face T_EMPTY font-lock-keyword-face T_HALT_COMPILER font-lock-keyword-face T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG font-lock-keyword-face T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG font-lock-keyword-face T_STATIC font-lock-keyword-face T_ABSTRACT font-lock-keyword-face T_FINAL font-lock-keyword-face T_PRIVATE font-lock-keyword-face T_PROTECTED font-lock-keyword-face T_PUBLIC font-lock-keyword-face T_UNSET font-lock-keyword-face T_LIST font-lock-keyword-face T_ARRAY font-lock-keyword-face T_CALLABLE font-lock-keyword-face T_STRING font-lock-string-face T_CONSTANT_ENCAPSED_STRING font-lock-string-face T_ENCAPSED_AND_WHITESPACE font-lock-string-face T_NUM_STRING font-lock-string-face T_DNUMBER font-lock-string-face T_LNUMBER font-lock-string-face T_INLINE_HTML font-lock-preprocessor-face T_ATTRIBUTE font-lock-doc-face T_DOC_COMMENT font-lock-doc-face T_COMMENT font-lock-doc-face END font-lock-doc-face T_VARIABLE font-lock-variable-name-face T_STRING_VARNAME font-lock-variable-name-face T_NAME_RELATIVE font-lock-variable-name-face T_NAME_QUALIFIED font-lock-variable-name-face T_NAME_FULLY_QUALIFIED font-lock-variable-name-face T_ENUM font-lock-keyword-face T_READONLY font-lock-keyword-face))
  "Syntax color table for tokens")

(defvar
  phps-mode-syntax-color--token-for-bookkeeping-p
  #s(hash-table size 65 test equal rehash-size 1.5 rehash-threshold 0.8125 data (T_VARIABLE t T_STRING t))
  "Flags whether token is used for bookkeeping or not.")


(provide 'phps-mode-syntax-color)
;;; phps-mode-syntax-color.el ends here
