;; Generated package description from phps-mode.el  -*- no-byte-compile: t -*-
(define-package "phps-mode" "0.4.40.0.20230223.210734" "Major mode for PHP with code intelligence" '((emacs "26")) :commit "e9be3ad876d0b81dcd75031995bab3a585507cad" :authors '(("Christian Johansson" . "christian@cvj.se")) :maintainer '("Christian Johansson" . "christian@cvj.se") :keywords '("tools" "convenience") :url "https://github.com/cjohansson/emacs-phps-mode")
